/* fio.c: File IO */
/* Copyright (c) 2000-2003 Fabien Tassin <fta@sofaraway.org> */

#ifdef MALLOC
# include <dmalloc.h>
#endif
#include <sys/time.h>
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include "fio.h"

FIO *fio_init(int fh) {
  FIO *fio;

  fio = (FIO *) malloc(sizeof(struct _FIO));
  fio->fh = fh;
  fio->size = FIO_BUFFER;
  fio->buffer = (char *) malloc(fio->size);
  fio->start = fio->end = fio->buffer;
  fio->status = FIO_OK;
  fio->debug = 0;
  return(fio);
}

void fio_free(FIO *fio) {
  free(fio->buffer);
  free(fio);
}

/* read a line in the given filehandle. Remove trailing '\n' or '\r\n' */
char *fio_readline(FIO *fio) {
  char *p, *q, *old;
  ssize_t sz;

  while (1) {
    if (fio->end > fio->start) {
      /* the buffer is not empty. Look for the first newline */
      p = memchr(fio->start, '\n', fio->end - fio->start);
      if (p != NULL) {
	/* found */
	int f = 0;
	if (p > fio->start && *(p - 1) == '\r')
 	  f++, p--;
	*p = '\0';
	old = fio->start;
	fio->start = p + (f ? 2 : 1);
	fio->status = FIO_OK;
	if (fio->debug > 0)
	  fprintf(stderr, "<<< \"%s%s\\n\"\n", old, f ? "\\r" : "");
	return(old);
      }
      if (fio->start != fio->buffer) {
	/* move unused data at the beginning of the buffer */
	for (p = fio->buffer, q = fio->start; q < fio->end; )
	  *p++ = *q++;
	fio->start = fio->buffer;
	fio->end = p;
      }
      p = fio->end;
    }
    else
      p = fio->buffer;

    fio->status = FIO_OK;
    /* read new data from the filehandle */
    sz = read(fio->fh, p, fio->size - (fio->end - fio->start));
    if (sz < 0) {
      perror("read() failed");
      fio->status = FIO_ERROR;
      return(NULL);
    }
    if (sz == 0)
      return(NULL); /* nothing to read */

    fio->start = fio->buffer;
    fio->end = &p[sz];

    /* Look for the first newline */
    p = memchr(fio->start, '\n', fio->end - fio->start);
    if (p != NULL) {
      /* found */
      int f = 0;
      if (p > fio->start && *(p - 1) == '\r')
	f++, p--;
      *p = '\0';
      old = fio->start;
      fio->start = p + (f ? 2 : 1);
      fio->status = FIO_OK;
      if (fio->debug > 0)
	fprintf(stderr, "<<< \"%s%s\\n\"\n", old, f ? "\\r" : "");
      return(old);
    }
    if ((fio->end - fio->start) >= fio->size) {
      /* not found but the buffer is full. This line is too long */
      fio->status = FIO_TOOLONG;
      fio->start = fio->end;
      return(NULL);
    }
    /* loop */
  }
}

/* write the buffer into the filehandle (no '\r\n' added) */
int fio_write(FIO *fio, char* buf) {
  int len = strlen(buf);
  if (fio->debug)
    fprintf(stderr, ">>> '%s'\n", buf);
  return(write(fio->fh, buf, len) == len);
}

/* write the buffer into the filehandle ('\r\n' added) */
int fio_writeline(FIO *fio, char* buf) {
  int r, old;
  char *s;

  s = (char *) malloc(strlen(buf) + 2);
  strcpy(s, buf);
  if (fio->debug)
    fprintf(stderr, ">>> \"%s\\r\\n\"\n", buf);
  old = fio->debug;
  fio->debug = 0;
  r = fio_write(fio, strcat(s, "\r\n"));
  fio->debug = old;
  free(s);
  return(r);
}
